import { ReturnModelType } from '@typegoose/typegoose';
import { AnyParamConstructor } from '@typegoose/typegoose/lib/types';
import { Application, Context } from '@midwayjs/koa';
import { cloneDeep, get } from 'lodash';
import { PageOptions, ResultOptions } from './Options';
import { PopulateOptions } from 'mongoose';
import { App, Inject } from '@midwayjs/core';
import { ServiceError } from '../error/CustomerError.error';
import { ErrorCode } from '../error/Codes';
import { GetModel } from './Utils';
/**
 * Service基类,实现了一些基础的crud
 */
export abstract class BaseService<T extends AnyParamConstructor<any>> {
  @App()
  app: Application;
  @Inject()
  ctx: Context;

  // 要求继承基类的service,必须给上默认的model
  abstract model: ReturnModelType<T>;
  /**
   * 带总数查询
   * @param {Object} filter 查询条件
   * @param {object} pageOptions 分页条件
   * @param {Boolean} lean 是否使用JavaScript形式数据;false为mongoose的模型实例数据
   * @param {Boolean} populate 是否进行ref关联数据
   * @returns {Promise<object>} 返回列表
   */
  async page(
    filter: object,
    pageOptions: PageOptions = {},
    resultOptions: ResultOptions = { lean: true, populate: true }
  ) {
    const data = await this.query(filter, pageOptions, resultOptions);
    const total = await this.count(filter);
    return { data, total };
  }

  /**
   * 列表查询-无查询条件处理
   * @param {Object} filter 查询条件
   * @param {object} pageOptions 分页条件
   * @param {Boolean} lean 是否使用JavaScript形式数据;false为mongoose的模型实例数据
   * @param {Boolean} populate 是否进行ref关联数据
   * @returns {Promise<object>} 返回列表
   */
  async query(
    filter: object,
    pageOptions: PageOptions = {},
    resultOptions: ResultOptions = { lean: true, populate: true }
  ): Promise<Array<any>> {
    const dup = cloneDeep(filter);
    const { lean, populate } = resultOptions;
    let refs = [];
    if (populate) refs = this.getRefs();
    const data = await this.model
      .find(dup, {}, { ...pageOptions })
      .populate(refs)
      .lean(lean);
    return data;
  }

  /**
   * 数据总数查询
   * @param {Object} filter 查询条件
   * @returns {number} 数据总数
   */
  async count(filter: Object): Promise<number> {
    const total = await this.model.count(filter);
    return total;
  }

  /**
   * 单查询-通过任意条件
   * @param query 查询条件
   * @param resultOptions 结果处理
   */
  async findOne(
    query: object = {},
    resultOptions: ResultOptions = { lean: true, populate: true }
  ): Promise<object | undefined> {
    const { lean, populate } = resultOptions;
    let refs = [];
    if (populate) refs = this.getRefs();
    const data = await this.model.findOne(query).populate(refs).lean(lean);
    return data;
  }

  /**
   * 修改
   * @param {object} filter 修改范围
   * @param {object} body 要修改的内容
   */
  async update(filter: object, body: object): Promise<void> {
    const filterKeys = Object.keys(filter);
    if (filterKeys.length <= 0) throw new ServiceError(ErrorCode.NEED_PARAMS);
    const num = await this.model.count(filter);
    if (num <= 0) throw new ServiceError(ErrorCode.DATA_NOT_FOUND);
    await this.model.updateMany(filter, body);
  }

  /**
   * 删除
   * @param {object} filter 要删除的数据范围
   *
   */
  async delete(filter: object): Promise<void> {
    const filterKeys = Object.keys(filter);
    if (filterKeys.length <= 0) throw new ServiceError(ErrorCode.NEED_PARAMS);
    await this.model.deleteMany(filter);
  }

  /**
   * 单创建
   * @param body 要创建的数据内容
   * @returns {object}
   */
  async create(body: object): Promise<object> {
    const data = await this.model.create(body);
    return data;
  }

  /**
   * 多创建
   * @param body 要创建的多个数据
   * @returns {object[]}
   */
  async createMany(body: object[]): Promise<Array<object>> {
    const data = await this.model.insertMany(body);
    return data;
  }
  /**
   * 获取本服务默认表的ref关系
   */
  getRefs(): Array<PopulateOptions> {
    const schema: any = get(this.model, 'schema.tree');
    const refs = [];
    for (const key in schema) {
      const f = schema[key];
      const ref = get(f, 'ref');
      if (ref) {
        const model = GetModel(ref);
        const path = key;
        refs.push({ path, model });
      }
    }
    return refs;
  }
}
