import { ReturnModelType } from '@typegoose/typegoose';
import { LoginRecord } from '../../entity/frame/loginRecord.entity';
import { Config, Inject, Provide } from '@midwayjs/core';
import { InjectEntityModel } from '@midwayjs/typegoose';
import { Context } from '@midwayjs/koa';
import { JwtService } from '@midwayjs/jwt';
import { get } from 'lodash';
import * as dayjs from 'dayjs';
import { ServiceError } from '../../error/CustomerError.error';
import { ErrorCode } from '../../error/Codes';

type modelType = ReturnModelType<typeof LoginRecord>;

@Provide()
export class LoginRecordService {
  @InjectEntityModel(LoginRecord)
  model: ReturnModelType<modelType>;
  @Inject()
  jwtService: JwtService;
  @Config('jwt.expiresIn')
  expire_step: number;
  @Inject()
  ctx: Context;

  timeFormat = 'YYYY-MM-DD HH:mm:ss';
  /**创建登录记录 */
  async create(token: string) {
    const userInfo = this.jwtService.decode(token);
    if (!userInfo) return;
    const user_id = get(userInfo, 'id', get(userInfo, '_id'));
    if (!user_id) return;
    const last_time = dayjs().format(this.timeFormat);
    const expire_time = dayjs()
      .add(this.expire_step, 's')
      .format(this.timeFormat);
    const req = this.ctx.request;
    const last_ip = get(req, 'header.x-forwarded-for', req.ip);
    const data = { token, last_time, expire_time, last_ip };
    const num = await this.model.count({ user_id });
    if (num > 0) {
      // 有过登录记录,那就直接更新token,last_time,expire_time,last_ip就行
      await this.model.updateOne({ user_id }, data);
    } else {
      // 没有记录,那就创建
      data['user_id'] = user_id;
      await this.model.create(data);
    }
  }
  /**更新过期时间 */
  async renewal(token) {
    const userInfo = this.jwtService.decode(token);
    if (!userInfo) return;
    const user_id = get(userInfo, 'id', get(userInfo, '_id'));
    if (!user_id) return;
    const data = await this.model.findOne({ user_id });
    if (!data) throw new ServiceError(ErrorCode.NOT_LOGIN);
    const req = this.ctx.request;
    const last_ip = get(req, 'header.x-forwarded-for', req.ip);
    const last_time = dayjs().format(this.timeFormat);
    const expire_time = dayjs()
      .add(this.expire_step, 's')
      .format(this.timeFormat);
    await this.model.updateOne(
      { user_id },
      { last_ip, last_time, expire_time }
    );
  }
  /**检查是否过期 */
  async check(token) {
    const userInfo = this.jwtService.decode(token);
    if (!userInfo) return;
    const user_id = get(userInfo, 'id', get(userInfo, '_id'));
    if (!user_id) return;
    const data = await this.model.findOne({ user_id });
    if (!data) throw new ServiceError(ErrorCode.NOT_LOGIN);
    const utoken = get(data, 'token');
    if (utoken === null) {
      // 抛出异常:用户已注销
      throw new ServiceError(ErrorCode.IS_LOGOUT);
    }
    if (utoken !== token) {
      // 抛出异常:用户已在其他地点登录
      throw new ServiceError(ErrorCode.OTHER_PLACE_LOGIN);
    }
    const expire_time = get(data, 'expire_time');
    const isBefore = dayjs().isBefore(expire_time);
    if (!isBefore) {
      // 抛出异常:登录已过期
      throw new ServiceError(ErrorCode.IS_EXPIRE);
    }
    return true;
  }
}
